@@ -61,7 +61,7 @@ class MultiClassClassifierBunchTests(unittest.TestCase):
 
     def test_metrics(self):
         self.assertIsInstance(self.explainer.metrics(), dict)
-        self.assertIsInstance(self.explainer.metrics_markdown(), str)
+        self.assertIsInstance(self.explainer.metrics_descriptions(), dict)
 
     def test_mean_abs_shap_df(self):
         self.assertIsInstance(self.explainer.mean_abs_shap_df(), pd.DataFrame)
@@ -112,9 +112,9 @@ class MultiClassClassifierBunchTests(unittest.TestCase):
 
     def test_shap_interaction_values_by_col(self):
         self.assertIsInstance(self.explainer.shap_interaction_values_by_col("Age"), np.ndarray)
-        self.assertEquals(self.explainer.shap_interaction_values_by_col("Age").shape, 
+        self.assertEqual(self.explainer.shap_interaction_values_by_col("Age").shape, 
                         self.explainer.shap_values.shape)
-        self.assertEquals(self.explainer.shap_interaction_values_by_col("Age", cats=True).shape, 
+        self.assertEqual(self.explainer.shap_interaction_values_by_col("Age", cats=True).shape, 
                         self.explainer.shap_values_cats.shape)
 
     def test_pdp_result(self):
@@ -247,8 +247,8 @@ class MultiClassClassifierBunchTests(unittest.TestCase):
         self.explainer.pos_label = "Southampton"
         self.assertIsInstance(self.explainer.pos_label, int)
         self.assertIsInstance(self.explainer.pos_label_str, str)
-        self.assertEquals(self.explainer.pos_label, 1)
-        self.assertEquals(self.explainer.pos_label_str, "Southampton")
+        self.assertEqual(self.explainer.pos_label, 1)
+        self.assertEqual(self.explainer.pos_label_str, "Southampton")
 
     def test_get_prop_for_label(self):
         self.explainer.pos_label = 1
